/**
 * Interface for GDB to communicate with c66x DSP
 *
 * Copyright (C) 2014 Texas Instruments, Incorporated
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation version 2.
 *
 * This program is distributed "as is" WITHOUT ANY WARRANTY of any kind,
 * whether express or implied; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <linux/kernel.h>
#include <linux/dma-mapping.h>
#include <linux/timer.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/mutex.h>
#include <linux/version.h>

#define GDB_KM_MAJOR_VERSION 1
#define GDB_KM_MINOR_VERSION 0                        
#define GDB_KM_PATCH_VERSION 1                      
#define GDB_TTY_MAJOR	     240                 
#define DRIVER_NAME	     "gdbserverproxy_module_drv"
#define NAME		     "gdbtty"
#define EDMA0_INT1	     0x02701010
#define EDMA0_INT2	     0x02701014
#define EDMA1_INT1	     0x02721010
#define EDMA1_INT2	     0x02721014
#define EDMA2_INT1	     0x02741010
#define EDMA2_INT2	     0x02741014
#define EDMA3_INT1	     0x02729010
#define EDMA3_INT2	     0x02729014
#define EDMA4_INT1	     0x02709010
#define EDMA4_INT2	     0x02709014
#define GDB_STR_SIZE	     544

/**
 * The third word of the shared data block will be encoded as follows
 * {detach command (1 byte), count of block (2 bytes), ready flag (1 byte)}
 * where ready flag {MSN(ibble) signals DSP and LSN signals the ARM)}
 */
#define READY_MASK                 0x00000001
#define COUNT_MASK                 0x00FFFF00
#define STATUS_MASK                0x000000FE

#define READY_FLAG                 0x00000001
#define STATUS_NEW_PACKET          0x00000010
#define DETACH_COMM                0x10000000

#define COUNT_SHIFT                8
#define ACK                        0xFFFFFFFF
#define JTAG_ID                    0x02620018
#define EDMA_CHAN_MASK             0x000000FF
#define EDMA_CC_MASK               0x0000FF00
#define DEBUG                      0
#if DEBUG
     #define POLL_DELAY_MS         500
#else
     #define POLL_DELAY_MS         10
#endif


/**
 * Check the kernel version and don't allow build if it doesn't agree
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,10,10)
#error "Version of linux kernel must be greater than 3.10.10"
#endif

/**
 * Handling resending of messages
 */ 
static char lastMsg[GDB_STR_SIZE];
static int  lastMsgCnt = 0;

struct dsp_session
{
     struct tty_port port;
     bool open;
     bool open_wronly;
     int openCount;
     struct mutex open_count_mutex;
     unsigned int sharedAddr;
     volatile int *handleSharedAddr;
};

/**
 * Structure used for storing session data
 */
struct tty_gdb_port 
{
     struct mutex port_write_mutex;
     unsigned int edmaChannel;
     volatile int *handleInt;
     int numberOfDsps;
     int openDevices;
};

/**
 *   Structure used for identifying keystone II device
 */
struct dsp_type_t
{
     char name[9];
     unsigned int jtag_id;
     unsigned int num_dsps;
};

static struct dsp_type_t dsp_type_table[] = {
    {.name = "TCI6614", .jtag_id = 0x0b96202f, .num_dsps = 4},
    {.name = "TCI6638", .jtag_id = 0x0b98102f, .num_dsps = 8},
    {.name = "TCI6630", .jtag_id = 0x0b9A702f, .num_dsps = 4},
    {.name = "66AK2Exx",.jtag_id = 0x0b9A602f, .num_dsps = 1},
    {.name = "Unknown", .jtag_id = 0xffffffff, .num_dsps = 8}  
};

static struct tty_gdb_port gdb_port;
struct dsp_session *dsp_data;

/**
 * timer used for polling
 */
static struct timer_list timer;

static int initializeDSP(int numDSP, struct dsp_session *dsp, struct device *dev);
static void polling_register(unsigned long data);
static struct dsp_type_t *get_dspinfo(void);

/**
 * FUNCTION NAME: triggerInt()
 *
 * DESCRIPTION: This function is to trigger an interrupt to the DSP
 *
 * ARGUMENTS:
 *           int  edmaChan: The EDMA channel to use for 
 *                          triggering the interrupt
 *           int handleInt: The pointer to the EDMA event set
 *                          set register (ESR)
 *
 */
static void triggerInt(unsigned int edmaChan, volatile int *handleInt, unsigned int status, volatile int *handleAddr)
{
     int writeVal = 0;

     if (edmaChan > 31)
          writeVal = 0x1 << (edmaChan - 32);
     else
          writeVal = 0x1 << edmaChan;

     // also write the ready bit in the status word
     if (handleAddr != NULL)
          iowrite32(status | STATUS_NEW_PACKET, handleAddr);

     if (handleInt != NULL)
          iowrite32(writeVal, handleInt);

}

/**
 * FUNCTION NAME: gdb_sendMessage()
 *
 * DESCRIPTION: This function sends a message to the DSP
 *
 * ARGUMENTS:
 *           const unsigned char *buf: The pointer to the message to be sent
 *           int                count: The count (in characters) of the 
 *                                     message to be sent
 *
 */
static int gdb_sendMessage(const unsigned char *buf, int count, struct dsp_session * dsp)
{
     if (buf != NULL && dsp->handleSharedAddr != NULL) {

          if (*buf == '-') {

               memcpy_toio(((char *)dsp->handleSharedAddr) + 12, lastMsg, lastMsgCnt);
               triggerInt(gdb_port.edmaChannel, gdb_port.handleInt, 0, dsp->handleSharedAddr + 2);
               if (DEBUG) {

                    printk("Message Sent= %.*s\n", lastMsgCnt, lastMsg);
                    printk("Sent Count= %d\n", lastMsgCnt);
               }
          }
          else if ((*buf != '+') && (*buf != '-')) {

               memcpy(lastMsg, buf, count);
               lastMsgCnt = count;
               memcpy_toio(((char *)dsp->handleSharedAddr) + 12, buf, count);
               triggerInt(gdb_port.edmaChannel, gdb_port.handleInt, 0, dsp->handleSharedAddr + 2);
               tty_insert_flip_char(&dsp->port, '+', 1);
               tty_flip_buffer_push(&dsp->port);
               if(DEBUG) {

                    printk("Message Sent= %.*s\n", count, buf);
                    printk("Sent Count= %d\n", count);
               }
          }
     }
     else
          count = 0;

	return count;
}

/**
 * FUNCTION NAME: gdb_open()
 *
 * DESCRIPTION: TTY operations open function
 *
 * ARGUMENTS:
 *           struct tty_struct *tty: The tty struct pointer
 *           struct file      *filp: File pointer
 * 
 *
 */
static int gdb_open(struct tty_struct *tty, struct file *filp)
{
     int ret = -ECONNREFUSED;
     int numDSP = tty->index;

     mutex_lock(&dsp_data[numDSP].open_count_mutex);
     dsp_data[numDSP].openCount++;
     mutex_unlock(&dsp_data[numDSP].open_count_mutex);

     // pass the specific dsp to read/write functions
     tty->driver_data = &dsp_data[numDSP];

     if (dsp_data[numDSP].open == true)
          return -EBUSY;

     // open with this flag to begin communication
     if (filp->f_flags & O_RDWR) {

          // only one dsp can get initialized at a time
          mutex_lock(&gdb_port.port_write_mutex);
          ret = initializeDSP(numDSP, &dsp_data[numDSP], tty->dev);
          if(ret == 0)
               gdb_port.openDevices++;
          mutex_unlock(&gdb_port.port_write_mutex);

          if (ret < 0) {

               if(DEBUG)
                    printk(KERN_ERR "Couldn't register gdbtty driver: Memory initialization failed\n");

               return ret;
          }

          if (gdb_port.openDevices == 1) {

               ret = mod_timer(&timer, jiffies + msecs_to_jiffies(POLL_DELAY_MS));
               if (ret < 0) {

                    if(DEBUG)
                         printk(KERN_ERR "Couldn't register gdbtty driver: Timer initialization failed\n");

                    return ret;
               }
          }

          return tty_port_open(&dsp_data[numDSP].port, tty, filp);
     }

     // open with this flag to write the initial address
     else if (filp->f_flags & O_WRONLY) {

          dsp_data[numDSP].open_wronly = true;
          dsp_data[numDSP].open = true;
          // pass the specific dsp to read/write functions
          tty->driver_data = &dsp_data[numDSP];
          return 0;
     }

     return ret;
     
}

/**
 * FUNCTION NAME: gdb_close()
 *
 * DESCRIPTION: TTY operations close function
 *
 * ARGUMENTS:
 *           struct tty_struct *tty: The tty struct pointer
 *           struct file      *filp: File pointer
 * 
 *
 */
static void gdb_close(struct tty_struct *tty, struct file *filp)
{
     struct dsp_session *dsp = tty->driver_data;

     if (NULL != dsp) {
          // trigger the interrupt, send release status, and release shared address
          if (dsp->open && (dsp->openCount == 1) && !dsp->open_wronly) {
 
               triggerInt(gdb_port.edmaChannel, gdb_port.handleInt, DETACH_COMM, dsp->handleSharedAddr + 2);
               release_mem_region(dsp->sharedAddr, GDB_STR_SIZE);
               iounmap(dsp->handleSharedAddr);
               dsp->open = false;
               tty_port_close(&dsp->port, tty, filp);

               mutex_lock(&gdb_port.port_write_mutex);
               gdb_port.openDevices--;
               mutex_unlock(&gdb_port.port_write_mutex);

               // this signifies the last device closing
               if (gdb_port.openDevices == 0) {
                    // release the timer
                    del_timer(&timer);

                    // unmap the interrupt and set the address to NULL
                    if(gdb_port.handleInt != NULL) {

                         iounmap(gdb_port.handleInt);
                         gdb_port.handleInt = NULL;
                    }
               }
          }
     
          if (dsp->open && dsp->open_wronly) {
               dsp->open = false;
               dsp->open_wronly = false;
          }

          mutex_lock(&dsp->open_count_mutex);
          dsp->openCount--;
          mutex_unlock(&dsp->open_count_mutex);
     }   
}

/**
 * FUNCTION NAME: gdb_write()
 *
 * DESCRIPTION: TTY operations write function
 *
 * ARGUMENTS:
 *           struct tty_struct *tty: The tty struct pointer
 *           struct file      *filp: File pointer
 * 
 *
 */
static int gdb_write(struct tty_struct *tty,
		const unsigned char *buf, int count)
{
     struct dsp_session *dsp = tty->driver_data;
     int ret = 0;
     unsigned int sharedAddress = 0;

     if ((NULL != dsp) && !dsp->open_wronly) {

	     mutex_lock(&gdb_port.port_write_mutex);
	     ret = gdb_sendMessage(buf, count, dsp);
	     mutex_unlock(&gdb_port.port_write_mutex);
     }
     
     if ((NULL != dsp) && dsp->open_wronly) {

          sscanf(buf, "%x", &sharedAddress);
          // the buffer is indexed by DSP number
          dsp->sharedAddr = sharedAddress + (tty->index * GDB_STR_SIZE);
          ret = count;
          if(DEBUG) {

               printk("opened file /dev/gdbtty%d\n", tty->index);
               printk("shared address = %x\n", dsp->sharedAddr);
          }
     }

     

	return ret;
}

/**
 * FUNCTION NAME: gdb_write_room()
 *
 * DESCRIPTION: TTY operations write_room function
 *
 * ARGUMENTS:
 *           struct tty_struct *tty: The tty struct pointer
 *           struct file      *filp: File pointer
 * 
 *
 */
static int gdb_write_room(struct tty_struct *tty)
{
     return GDB_STR_SIZE;
}

/**
 * FUNCTION NAME: gdb_ioctl()
 *
 * DESCRIPTION: TTY operations ioctl function
 *
 * ARGUMENTS:
 *           struct tty_struct *tty: The tty struct pointer
 *           struct file      *filp: File pointer
 * 
 *
 */
static int gdb_ioctl(struct tty_struct *tty,
			unsigned int cmd, unsigned long arg)
{
     struct dsp_session *dsp = tty->driver_data;

     if (NULL == dsp)
          return -EINVAL;

     return -ENOIOCTLCMD;
}

/**
 * Function pointer for tty operations
 */
static const struct tty_operations tty_gdb_ops = {
     .open          = gdb_open,
     .close         = gdb_close,
     .write         = gdb_write,
     .write_room    = gdb_write_room,
     .ioctl         = gdb_ioctl,
};

 
/**
 * FUNCTION NAME: initializeDSP()
 *
 * DESCRIPTION: Initialize the DSP to be debugged and 
 *              allocate the necessary memory regions
 *
 * ARGUMENTS:
 *           int              numDSP: The DSP number
 *           struct dsp_session *dsp: Pointer to the current dsp_session
 *           struct device      *dev: Pointer to the device
 *
 */
static int initializeDSP(int numDSP, struct dsp_session *dsp, struct device *dev)
{
     int retries = 25;
     int edmaInfo = 0;
     int edmaCC = 0;
     int error = 0;
     int edmaInt1 = 0;
     int edmaInt2 = 0;

     if(dsp->sharedAddr != 0) {

          if(request_mem_region(dsp->sharedAddr, GDB_STR_SIZE, DRIVER_NAME) == NULL) {

               if (DEBUG)
                    printk("Request for memory failed\n");
               return -EADDRNOTAVAIL;
          }
          // map the shared memory
          dsp->handleSharedAddr = ioremap_nocache((volatile unsigned int)dsp->sharedAddr, GDB_STR_SIZE);
     }
     else {

          if(DEBUG)
               printk("Shared address is uninitialized\n");
          return -ECONNREFUSED;
     }

     // get the edma channel
     if (dsp->handleSharedAddr != NULL) {

          edmaInfo = ioread32(dsp->handleSharedAddr + 1);
          // extract the channel controller and channel numbers
          gdb_port.edmaChannel = EDMA_CHAN_MASK & edmaInfo;
          edmaCC = (EDMA_CC_MASK & edmaInfo) >> 8;

          if(DEBUG) {

               printk("EDMA Controller = %d\n", edmaCC);
               printk("EDMA Channel = %d\n", gdb_port.edmaChannel);
          }

          switch(edmaCC) {
               case 1:
                    edmaInt1 = EDMA1_INT1;
                    edmaInt2 = EDMA1_INT2;
                    break;
               case 2:
                    edmaInt1 = EDMA2_INT1;
                    edmaInt2 = EDMA2_INT2;
                    break;
               case 3:
                    edmaInt1 = EDMA3_INT1;
                    edmaInt2 = EDMA3_INT2;
                    break;
               case 4:
                    edmaInt1 = EDMA4_INT1;
                    edmaInt2 = EDMA4_INT2;
                    break;
               // EDMACC 0 is used by the kernel
               case 0:
               default:
                    if (DEBUG)
                         printk("EDMA channel controller %d is invalid\n", edmaCC);
                    goto fail_1;
          }

          // if this isn't NULL, we aren't the first device to map the register, so
          // just use the current (valid) handle
          if(gdb_port.handleInt == NULL) {

               if ((gdb_port.edmaChannel > 31) && (gdb_port.edmaChannel < 64))
                    gdb_port.handleInt = ioremap_nocache((volatile unsigned int)edmaInt2, 4);
               else if (gdb_port.edmaChannel < 32)
                    gdb_port.handleInt = ioremap_nocache((volatile unsigned int)edmaInt1, 4);
               else {

                    if (DEBUG)
                         printk("EDMA channel is invalid\n");
                    goto fail_1;
               }
          }

          // initialize the DSP
          if (gdb_port.handleInt != NULL) {

               // write the DSP num (+1) to initialize the specific DSP
               iowrite32(numDSP + 1, dsp->handleSharedAddr);
               triggerInt(gdb_port.edmaChannel, gdb_port.handleInt, 0, NULL);
          }
          else {

               if (DEBUG)
                    printk("Couldn't allocate interrupt register\n");
               goto fail_2;
          }

          // wait for response
          while ((ioread32(dsp->handleSharedAddr) != ACK) && (retries > 0))
               retries--;


          if (retries <= 0) {

               if(DEBUG)
                    printk("The DSP did not respond to the request\n");
               goto fail_2;
          }

          dsp->open = true;
          if (DEBUG) {

               printk("Physical address = %p\n", (void *)dsp->sharedAddr);
               printk("Virtual address = %p\n", (void *)dsp->handleSharedAddr);
          }
          
          return error;
     }  

fail_2:
     if (gdb_port.handleInt != NULL) {

          iounmap(gdb_port.handleInt);
          gdb_port.handleInt = NULL;
     }

fail_1:
     if (dsp->handleSharedAddr != NULL) {

          release_mem_region(dsp->sharedAddr, GDB_STR_SIZE);
          iounmap(dsp->handleSharedAddr);
     }

     
     return -ECONNREFUSED;
}

/**
 * FUNCTION NAME: polling_register()
 *
 * DESCRIPTION: Continually polls a memory region in order
 *              to read data sent from the DSP
 *
 */
static void polling_register(unsigned long data)
{
     int value = 0;
     char *message = 0;
     int count = 0;
     int device = 0;

     // iterate through all DSPs to ensure all messages are consumed
     for (device = 0; device < gdb_port.numberOfDsps; device++) {

          if (dsp_data[device].open == true) {

               // read the status to find out if a new message is available
               value = ioread32(dsp_data[device].handleSharedAddr + 2);

               // process the data
               if (READY_FLAG == (READY_FLAG & value)) {

                    count = ((value & COUNT_MASK) >> COUNT_SHIFT);
                    if ((count > 0) && (count <= GDB_STR_SIZE)) {

                         message = (char *) kmalloc(count, GFP_KERNEL);
                         memcpy_fromio(message, ((char *)dsp_data[device].handleSharedAddr) + 12, count);
                         if(DEBUG) {

                              printk("Message received = %.*s\n", count, message);
                              printk("Received Count = %d\n", count);
                         }
                         tty_insert_flip_string(&dsp_data[device].port, message, count);
                         tty_flip_buffer_push(&dsp_data[device].port);
                         kfree(message);
                    }
                    else {
                         if (DEBUG)
                              printk("Count check failed = %d\n", count);
                    }

                    // clear size and ready bit
                    iowrite32(0, dsp_data[device].handleSharedAddr + 2);
               }
          }        
     }
     // reset the timer
     mod_timer(&timer, jiffies + msecs_to_jiffies(POLL_DELAY_MS));
}

/**
 * FUNCTION NAME: get_dspinfo()
 *
 * DESCRIPTION: provides number of DSPs in a device
 *
 */
static struct dsp_type_t *get_dspinfo(void)
{
     volatile int *pjtagID = 0;
     int jtagID = 0;
     int i = 0;
     // do not compare against "unknown" at the end of table
     int num_elements = sizeof(dsp_type_table)/sizeof(struct dsp_type_t) - 1;
     // map the jtag id register to current address space
     pjtagID = ioremap_nocache((volatile unsigned int)JTAG_ID, 4);

     if(pjtagID != NULL) {

          jtagID = ioread32(pjtagID);
          iounmap(pjtagID);
     }

     // compare to table of known IDs
     for (i = 0; i < num_elements; i++) {

          if (jtagID == dsp_type_table[i].jtag_id)
               break;
     }
     
     return &dsp_type_table[i];
}

static struct tty_port_operations null_ops = { };

static struct tty_driver *tty_gdb_driver;

static int __init gdb_dev_init(void)
{
     int ret = -ENOMEM;
     int device = 0;

     // find number of DSPs in device
     struct dsp_type_t *dspType = get_dspinfo();
     
     if (DEBUG)
          printk("number of DSPs = %d\n", dspType->num_dsps);
     
     // allocate space for number of DSPs
     dsp_data = (struct dsp_session *)kmalloc(dspType->num_dsps * sizeof(struct dsp_session), GFP_KERNEL);
     
     memset(dsp_data, 0, dspType->num_dsps * sizeof(struct dsp_session));

     mutex_init(&gdb_port.port_write_mutex);

     tty_gdb_driver = tty_alloc_driver(dspType->num_dsps,
               TTY_DRIVER_RESET_TERMIOS |
               TTY_DRIVER_REAL_RAW);
     if (IS_ERR(tty_gdb_driver))
          return PTR_ERR(tty_gdb_driver);

     // initialize the local data structure for all DSPs
     for (device = 0; device < dspType->num_dsps; device++) {

	     tty_port_init(&dsp_data[device].port);
	     dsp_data[device].port.ops = &null_ops;
          mutex_init(&dsp_data[device].open_count_mutex);
     }
     gdb_port.openDevices = 0;
     gdb_port.numberOfDsps = dspType->num_dsps;
     gdb_port.edmaChannel = 0;
     gdb_port.handleInt = NULL;

     tty_gdb_driver->driver_name = DRIVER_NAME;
     tty_gdb_driver->name = NAME;
     tty_gdb_driver->major = GDB_TTY_MAJOR;
     tty_gdb_driver->minor_start = 0;
     tty_gdb_driver->type = TTY_DRIVER_TYPE_CONSOLE;
     tty_gdb_driver->init_termios = tty_std_termios;
     tty_gdb_driver->init_termios.c_oflag = OPOST | OCRNL | ONOCR | ONLRET;
     tty_set_operations(tty_gdb_driver, &tty_gdb_ops);

     for (device = 0; device < gdb_port.numberOfDsps; device++)
          tty_port_link_device(&dsp_data[device].port, tty_gdb_driver, device);

     ret = tty_register_driver(tty_gdb_driver);

     if (ret < 0) {

          printk(KERN_ERR "Couldn't register gdbtty driver.\n");
          put_tty_driver(tty_gdb_driver);
          return ret;
     }

     // setup the timer for polling
     setup_timer(&timer, polling_register, 0);

     return 0;
}

static void __exit gdb_dev_exit(void)
{
     int device = 0;

     kfree(dsp_data);

     // unregister each device
     for(device = 0; device < gdb_port.numberOfDsps; device++)
          tty_unregister_device(tty_gdb_driver, device);

     // unregister driver
     tty_unregister_driver(tty_gdb_driver);
}

/* Module information */
MODULE_AUTHOR("Texas Instruments Incorporated");
MODULE_DESCRIPTION("Interface for GDB to communicate with c66x DSP");
MODULE_LICENSE("GPL v2");
MODULE_ALIAS_CHARDEV_MAJOR(GDB_TTY_MAJOR);

module_exit(gdb_dev_exit);
module_init(gdb_dev_init);;









